# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-24  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from contextlib import contextmanager
from datetime import datetime, timedelta
from http.server import BaseHTTPRequestHandler, HTTPServer
import io
import logging
import os
import re
import time
import urllib

import keyring
import PIL.Image as PIL
import requests

from photini import __version__
from photini.configstore import key_store
from photini.metadata import Metadata
from photini.pyqt import *
from photini.pyqt import using_pyside
from photini.widgets import Label, StartStopButton

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class UploadAborted(Exception):
    pass


class UploaderSession(QtCore.QObject):
    upload_progress = QtSignal(dict)
    new_token = QtSignal(dict)
    headers = {'User-Agent': 'Photini/' + __version__}

    def __init__(self, user_data={}, client_data={}, parent=None):
        super(UploaderSession, self).__init__(parent=parent)
        self.user_data = user_data
        self.client_data = client_data
        self.api = None
        self.open_connection()

    def open_connection(self):
        if not self.api:
            self.api = requests.Session()
            self.api.headers.update(self.headers)

    def close_connection(self):
        if self.api:
            self.api.close()
            self.api = None

    @staticmethod
    def check_response(rsp, decode=True):
        UploaderSession.check_interrupt()
        try:
            rsp.raise_for_status()
            if decode:
                rsp = rsp.json()
        except UploadAborted:
            raise
        except Exception as ex:
            logger.error(str(ex))
            return None
        return rsp

    @staticmethod
    def check_interrupt():
        if QtCore.QThread.currentThread().isInterruptionRequested():
            raise UploadAborted

    def progress(self, monitor):
        self.upload_progress.emit(
            {'value': monitor.bytes_read * 100 // monitor.len})

    def upload_files(self, upload_list):
        upload_count = 0
        for image, convert, params in upload_list:
            upload_count += 1
            self.upload_progress.emit({
                'label': '{} ({}/{})'.format(os.path.basename(image.path),
                                             upload_count, len(upload_list)),
                'busy': True})
            with self.open_file(image, convert) as (image_type, fileobj):
                error = self.do_upload(fileobj, image_type, image, params)
            if error:
                self.upload_progress.emit({'error': (image, error)})

    @contextmanager
    def open_file(self, image, convert):
        if convert:
            exiv_image, image_type = convert(image)
            exiv_io = exiv_image.io()
            fileobj = AbortableFileReader(io.BytesIO(exiv_io), exiv_io.size())
            try:
                yield image_type, fileobj
            finally:
                fileobj.close()
        else:
            fileobj = AbortableFileReader(
                open(image.path, 'rb'), os.stat(image.path).st_size)
            try:
                yield image.file_type, fileobj
            finally:
                fileobj.close()


class AbortableFileReader(object):
    def __init__(self, fileobj, size):
        super(AbortableFileReader, self).__init__()
        self._f = fileobj
        # requests library uses 'len' attribute instead of seeking to
        # end of file and back
        self.len = size

    # substitute read method
    def read(self, size):
        UploaderSession.check_interrupt()
        return self._f.read(size)

    # delegate all other attributes to file object
    def __getattr__(self, name):
        if name == 'getvalue':
            # don't allow requests library to suck in all data at once
            raise AttributeError(name)
        return getattr(self._f, name)


class UploadWorker(QtCore.QObject):
    finished = QtSignal()
    upload_progress = QtSignal(dict)

    def __init__(self, session, upload_list, *args, **kwds):
        super(UploadWorker, self).__init__(*args, **kwds)
        self.session = session
        self.upload_list = upload_list

    @QtSlot()
    @catch_all
    def start(self):
        with self.session(parent=self) as session:
            session.upload_progress.connect(self.upload_progress)
            try:
                session.upload_files(self.upload_list)
            except UploadAborted:
                pass
            except Exception as ex:
                logger.exception(ex)
        self.upload_progress.emit({'value': 0, 'label': None, 'busy': False})
        self.finished.emit()


class AuthRequestHandler(BaseHTTPRequestHandler):
    def log_message(self, format_, *args):
        logger.debug(format_, *args)

    @catch_all
    def do_GET(self):
        query = urllib.parse.urlsplit(self.path).query
        self.server.result = urllib.parse.parse_qs(query)
        logger.info('do_GET: %s', repr(self.server.result))
        title = translate('UploaderTabsAll', 'Close window')
        text = translate(
            'UploaderTabsAll', 'You may now close this browser window.')
        response = '''
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"
"http://www.w3.org/TR/html4/strict.dtd">
<html>
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8">
    <title>{title}</title>
  </head>
  <body>
    <h1>{title}</h1>
    <p>{text}</p>
  </body>
</html>
'''.format(title=title, text=text)
        response = response.encode('utf-8')
        self.send_response(200)
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", str(len(response)))
        self.end_headers()
        self.wfile.write(response)


class AuthServer(QtCore.QObject):
    finished = QtSignal()

    @QtSlot()
    @catch_all
    def handle_requests(self):
        self.running = True
        self.server.timeout = 10
        self.server.result = None
        while self.running:
            self.server.handle_request()
            if self.server.result:
                break
        self.server.server_close()
        self.finished.emit()


class UploaderUser(QtWidgets.QGridLayout):
    connection_changed = QtSignal(bool)

    def __init__(self, *arg, **kw):
        super(UploaderUser, self).__init__(*arg, **kw)
        self.setContentsMargins(0, 0, 0, 0)
        # dictionary of unavailable widgets (e.g. server version too low)
        self.unavailable = {}
        # user details
        group = QtWidgets.QGroupBox()
        group.setMinimumWidth(width_for_text(group, 'x' * 17))
        group.setLayout(QtWidgets.QVBoxLayout())
        self.user_photo = QtWidgets.QLabel()
        self.user_photo.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop)
        group.layout().addWidget(self.user_photo)
        self.user_name = QtWidgets.QLabel()
        self.user_name.setWordWrap(True)
        self.user_name.setMinimumWidth(10)
        group.layout().addWidget(self.user_name)
        group.layout().addStretch(1)
        self.addWidget(group, 0, 0)
        self.setRowStretch(0, 1)
        # connect / disconnect button
        self.connect_button = StartStopButton(
            translate('UploaderTabsAll', 'Log in'),
            translate('UploaderTabsAll', 'Log out'))
        self.connect_button.click_start.connect(self.log_in)
        self.connect_button.click_stop.connect(self.log_out)
        self.addWidget(self.connect_button, 1, 0)
        # other init
        self.client_data = {'name': self.config_section}
        if key_store.config.has_section(self.config_section):
            for option in key_store.config.options(self.config_section):
                self.client_data[option] = key_store.get(
                    self.config_section, option)
        self.user_data = {}

    @contextmanager
    def session(self, **kw):
        try:
            session = self.new_session(**kw)
            yield session
        finally:
            session.close_connection()

    def show_user(self, name, picture):
        if name:
            name = name[:10].replace(' ', '\xa0') + name[10:]
            self.user_name.setText(translate(
                'UploaderTabsAll', 'Logged in as {user} on {service}'
                ).format(user=name, service=self.service_name()))
        else:
            self.user_name.setText(translate(
                'UploaderTabsAll', 'Not logged in to {service}'
                ).format(service=self.service_name()))
        pixmap = QtGui.QPixmap()
        if picture:
            pixmap.loadFromData(picture)
            max_width = self.user_photo.frameRect().width()
            if pixmap.width() > max_width:
                pixmap = pixmap.scaledToWidth(
                    max_width, Qt.TransformationMode.SmoothTransformation)
        self.user_photo.setPixmap(pixmap)

    @QtSlot()
    @catch_all
    def log_out(self):
        if keyring.get_password('photini', self.config_section):
            self.unauthorise()
            keyring.delete_password('photini', self.config_section)
        self.user_data = {}
        self.connection_changed.emit(False)

    @QtSlot()
    @catch_all
    def log_in(self, do_auth=True):
        with DisableWidget(self.connect_button):
            if self.load_user_data():
                self.connection_changed.emit(True)
            elif do_auth:
                self.authorise()

    def unauthorise(self):
        pass

    def authorise(self):
        with Busy():
            # do full authentication procedure
            frob_or_uri = self.get_frob()
            if frob_or_uri:
                http_server = None
                auth_response = frob_or_uri
            else:
                # create temporary local web server
                http_server = HTTPServer(('127.0.0.1', 0), AuthRequestHandler)
                frob_or_uri = 'http://127.0.0.1:' + str(http_server.server_port)
            auth = self.auth_exchange(frob_or_uri)
            auth_url = next(auth)
            if not auth_url:
                self.logger.error('Failed to get auth URL')
                if http_server:
                    http_server.server_close()
                return
            if http_server:
                server = AuthServer()
                thread = QtCore.QThread(self)
                server.moveToThread(thread)
                server.server = http_server
                thread.started.connect(server.handle_requests)
                server.finished.connect(thread.quit)
                server.finished.connect(server.deleteLater)
                thread.finished.connect(thread.deleteLater)
                thread.start()
            if not QtGui.QDesktopServices.openUrl(QtCore.QUrl(auth_url)):
                self.logger.error('Failed to open web browser')
                return
        # wait for user to authorise in web browser
        dialog = QtWidgets.QMessageBox(self.parentWidget())
        dialog.setWindowTitle(
            translate('UploaderTabsAll', 'Photini: authorise'))
        dialog.setText('<h3>{}</h3>'.format(translate(
            'UploaderTabsAll', 'Authorisation required')))
        dialog.setInformativeText(translate(
            'UploaderTabsAll', 'Please use your web browser to authorise'
            ' Photini, and then close this dialog.'))
        dialog.setIcon(dialog.Icon.Warning)
        dialog.setStandardButtons(dialog.StandardButton.Ok)
        if http_server:
            server.finished.connect(dialog.close)
        execute(dialog)
        if http_server:
            auth_response = http_server.result
            server.running = False
        if not auth_response:
            return
        with Busy():
            try:
                auth.send(auth_response)
            except StopIteration:
                pass

    def get_frob(self):
        return None

    def get_password(self):
        return keyring.get_password('photini', self.config_section)

    def set_password(self, password):
        keyring.set_password('photini', self.config_section, password)


class AlbumList(QtWidgets.QWidget):
    def __init__(self, *arg, max_selected=0, **kw):
        super(AlbumList, self).__init__(*arg, **kw)
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setSpacing(0)
        self.layout().setSizeConstraint(
            QtWidgets.QLayout.SizeConstraint.SetMinAndMaxSize)
        self.album_widgets = []
        self.max_selected = max_selected
        self.checked_widgets = []

    def add_album(self, album, index=-1):
        widget = QtWidgets.QCheckBox(album['title'].replace('&', '&&'))
        if album['description']:
            widget.setToolTip('<p>' + album['description'] + '</p>')
        widget.setEnabled(album['writeable'])
        widget.setProperty('id', album['id'])
        if self.max_selected:
            widget.stateChanged.connect(self.state_changed)
        if index >= 0:
            self.layout().insertWidget(index, widget)
            self.album_widgets.insert(index, widget)
        else:
            self.layout().addWidget(widget)
            self.album_widgets.append(widget)
        return widget

    @QtSlot(int)
    @catch_all
    def state_changed(self, state):
        for widget in list(self.checked_widgets):
            if not widget.isChecked():
                self.checked_widgets.remove(widget)
        for widget in self.get_checked_widgets():
            if widget not in self.checked_widgets:
                self.checked_widgets.append(widget)
        while len(self.checked_widgets) > self.max_selected:
            widget = self.checked_widgets.pop(0)
            widget.setChecked(False)

    def clear_albums(self):
        for widget in self.album_widgets:
            self.layout().removeWidget(widget)
            widget.setParent(None)
        self.album_widgets = []

    def get_checked_widgets(self):
        result = []
        for widget in self.album_widgets:
            if widget.isChecked():
                result.append(widget)
        return result

    def get_checked_ids(self):
        return [x.property('id') for x in self.get_checked_widgets()]


class PhotiniUploader(QtWidgets.QWidget):
    def __init__(self, *arg, **kw):
        super(PhotiniUploader, self).__init__(*arg, **kw)
        self.app = QtWidgets.QApplication.instance()
        self.app.aboutToQuit.connect(self.shutdown)
        self.logger.debug('using %s', keyring.get_keyring().__module__)
        self.setLayout(QtWidgets.QGridLayout())
        self.upload_worker = None
        # dictionary of all widgets with parameter settings
        self.widget = {}
        # dictionary of control buttons
        self.buttons = {}
        ## first column is "user" widget
        self.user_widget.connection_changed.connect(self.connection_changed)
        self.layout().addLayout(self.user_widget, 0, 0)
        ## remaining columns are 'service' specific
        self.config_layouts = []
        column_count = 1
        for layout, stretch in self.config_columns():
            self.config_layouts.append(layout)
            self.layout().addLayout(layout, 0, column_count)
            self.layout().setColumnStretch(column_count, stretch)
            column_count += 1
        ## bottom row
        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.layout().addLayout(layout, 1, 0, 1, column_count)
        # progress bar
        self.progress_label = QtWidgets.QLabel()
        layout.addWidget(self.progress_label, 0, 0)
        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setFormat('%p%')
        layout.addWidget(self.progress_bar, 0, 1)
        layout.setColumnStretch(1, 1)
        self.upload_progress({'label': None, 'value': 0})
        # upload button
        self.buttons['upload'] = StartStopButton(
            translate('UploaderTabsAll', 'Start upload'),
            translate('UploaderTabsAll', 'Stop upload'))
        self.buttons['upload'].setEnabled(False)
        self.buttons['upload'].click_start.connect(self.start_upload)
        self.buttons['upload'].click_stop.connect(self.stop_upload)
        layout.addWidget(self.buttons['upload'], 0, 2)
        # initialise as not connected
        self.connection_changed(False)

    def album_list(self, label=None, max_selected=0):
        # list of albums widget
        column = QtWidgets.QGridLayout()
        column.setContentsMargins(0, 0, 0, 0)
        group = QtWidgets.QGroupBox()
        group.setMinimumWidth(width_for_text(group, 'x' * 23))
        group.setLayout(QtWidgets.QVBoxLayout())
        label = label or translate('UploaderTabsAll', 'Add to albums')
        group.layout().addWidget(QtWidgets.QLabel(label))
        scrollarea = QtWidgets.QScrollArea()
        scrollarea.setFrameStyle(scrollarea.Shape.NoFrame)
        self.widget['albums'] = AlbumList(max_selected=max_selected)
        scrollarea.setWidget(self.widget['albums'])
        group.layout().addWidget(scrollarea)
        column.addWidget(group, 0, 0)
        return column

    @QtSlot()
    @catch_all
    def shutdown(self):
        self.stop_upload()
        while self.upload_worker and self.upload_worker.thread().isRunning():
            self.app.processEvents()

    @QtSlot(bool)
    @catch_all
    def connection_changed(self, connected):
        if 'albums' in self.widget:
            self.widget['albums'].clear_albums()
        self.user_widget.show_user(None, None)
        if connected:
            with Busy():
                for key, value in self.user_widget.on_connect(self.widget):
                    if key == 'connected':
                        connected = value
                        if not connected:
                            break
                    elif key == 'user':
                        self.user_widget.show_user(*value)
                    elif key == 'album':
                        self.widget['albums'].add_album(value)
                    self.app.processEvents()
        self.user_widget.connect_button.set_checked(connected)
        self.enable_config(connected and not self.upload_worker)
        self.user_widget.connect_button.setEnabled(not self.upload_worker)
        self.enable_upload_button()

    def enable_config(self, enabled):
        for layout in self.config_layouts:
            for idx in range(layout.count()):
                widget = layout.itemAt(idx).widget()
                if widget:
                    widget.setEnabled(enabled)
        for key in self.user_widget.unavailable:
            self.widget[key].setEnabled(
                enabled and not self.user_widget.unavailable[key])

    def refresh(self):
        if not self.user_widget.connect_button.is_checked():
            self.user_widget.log_in(do_auth=False)
        self.enable_upload_button()

    def do_not_close(self):
        if not self.upload_worker:
            return False
        dialog = QtWidgets.QMessageBox(parent=self)
        dialog.setWindowTitle(
            translate('UploaderTabsAll', 'Photini: upload in progress'))
        dialog.setText('<h3>{}</h3>'.format(translate(
            'UploaderTabsAll',  'Upload to {service} has not finished.'
            ).format(service=self.user_widget.service_name())))
        dialog.setInformativeText(translate(
            'UploaderTabsAll', 'Closing now will terminate the upload.'))
        dialog.setIcon(dialog.Icon.Warning)
        dialog.setStandardButtons(
            dialog.StandardButton.Close | dialog.StandardButton.Cancel)
        dialog.setDefaultButton(dialog.StandardButton.Cancel)
        result = execute(dialog)
        return result == dialog.StandardButton.Cancel

    def read_image(self, image):
        with open(image.path, 'rb') as f:
            data = f.read()
        return {'image': None,
                'width': None,
                'height': None,
                'data': image.metadata.clone(data),
                'mime_type': image.file_type,
                'metadata': image.metadata,
                'name': os.path.basename(image.path),
                }

    def data_to_image(self, src):
        dst = dict(src)
        if dst['image']:
            return dst
        exiv_io = dst['data'].io()
        dst['image'] = PIL.open(io.BytesIO(exiv_io))
        dst['width'], dst['height'] = dst['image'].size
        return dst

    def image_to_data(self, src, mime_type=None, max_size=None):
        dst_mime_type = mime_type or src['mime_type']
        if src['data'] and src['mime_type'] == dst_mime_type and not (
                            max_size and src['data'].io().size() > max_size):
            return src
        src = self.data_to_image(src)
        w_src, h_src = src['width'], src['height']
        for scale in (1000, 680, 470, 330, 220, 150,
                      100, 68, 47, 33, 22, 15, 10, 7, 5, 3, 2, 1):
            dst = self.data_to_image(src)
            w_dst = w_src * scale // 1000
            h_dst = h_src * scale // 1000
            if scale != 1000:
                dst = self.resize_image(dst, w_dst, h_dst)
            dst_mime_type = mime_type or dst['mime_type']
            fmt = dst_mime_type.split('/')[1].upper()
            if dst_mime_type == 'image/jpeg':
                options = [{'quality': 95}, {'quality': 85}, {'quality': 75}]
            else:
                options = [{}]
            for option in options:
                dest_buf = io.BytesIO()
                dst['image'].save(dest_buf, format=fmt, **option)
                data = dest_buf.getbuffer()
                dst['data'] = src['metadata'].clone(data)
                dst['mime_type'] = dst_mime_type
                if not (max_size and dst['data'].io().size() > max_size):
                    logger.info('Converted %s from %dx%d to %dx%d JPEG',
                                src['name'], w_src, h_src, w_dst, h_dst)
                    return dst
        return None

    def resize_image(self, src, w, h):
        dst = self.data_to_image(src)
        dst['image'] = dst['image'].resize((w, h), resample=PIL.BICUBIC)
        dst['width'], dst['height'] = dst['image'].size
        dst['data'] = None
        dst['mime_type'] = 'image/jpeg'
        return dst

    def process_image(self, image):
        image = self.read_image(image)
        image = self.image_to_data(
            image, mime_type='image/jpeg',
            max_size=self.user_widget.max_size['image']['bytes'])
        return image['data'], image['mime_type']

    def copy_file_and_metadata(self, image):
        image = self.read_image(image)
        return image['data'], image['mime_type']

    def add_skip_button(self, dialog):
        return dialog.addButton(translate('UploaderTabsAll', 'Skip'),
                                dialog.ButtonRole.AcceptRole)

    def ask_resize_image(self, image, state, resizable=False, pixels=None):
        if 'ask_resize' not in state:
            state['ask_resize'] = True
        max_size = self.user_widget.max_size[image.file_type.split('/')[0]]
        if pixels:
            if 'pixels' not in max_size:
                return {}
            max_size = max_size['pixels']
            size = pixels
            text = translate(
                'UploaderTabsAll', 'File "{file_name}" has {size} pixels and'
                ' exceeds {service}\'s limit of {max_size} pixels.')
        else:
            max_size = max_size['bytes']
            size = os.stat(image.path).st_size
            text = translate(
                'UploaderTabsAll', 'File "{file_name}" has {size} bytes and'
                ' exceeds {service}\'s limit of {max_size} bytes.')
        if size <= max_size:
            return {}
        if resizable and not state['ask_resize']:
            return {'resize': True}
        dialog = QtWidgets.QMessageBox(parent=self)
        dialog.setWindowTitle(
            translate('UploaderTabsAll', 'Photini: too large'))
        dialog.setText('<h3>{}</h3>'.format(
            translate('UploaderTabsAll', 'File too large.')))
        text = text.format(
                file_name=os.path.basename(image.path), size=size,
                service=self.user_widget.service_name(), max_size=max_size)
        if resizable:
            text += ' ' + translate(
                'UploaderTabsAll', 'Would you like to resize it?')
            dialog.setStandardButtons(dialog.StandardButton.Yes |
                                      dialog.StandardButton.YesToAll)
        self.add_skip_button(dialog)
        dialog.setInformativeText(text)
        dialog.setIcon(dialog.Icon.Warning)
        result = execute(dialog)
        if result == dialog.StandardButton.Yes:
            return {'resize': True}
        if result == dialog.StandardButton.YesToAll:
            state['ask_resize'] = False
            return {'resize': True}
        return {'omit': True}

    def ask_convert_image(self, image, convertible=True, mime_type=None):
        mime_type = mime_type or image.file_type
        dialog = QtWidgets.QMessageBox(parent=self)
        dialog.setWindowTitle(translate(
            'UploaderTabsAll', 'Photini: incompatible type'))
        dialog.setText('<h3>{}</h3>'.format(translate(
            'UploaderTabsAll', 'Incompatible image type.')))
        dialog.setStandardButtons(dialog.StandardButton.Yes)
        dialog.setDefaultButton(dialog.StandardButton.Yes)
        self.add_skip_button(dialog)
        text = translate(
            'UploaderTabsAll', 'File "{file_name}" is of type "{file_type}",'
            ' which {service} may not handle correctly.'
            ).format(file_name=os.path.basename(image.path),
                     file_type=mime_type,
                     service=self.user_widget.service_name())
        if convertible:
            text += ' ' + translate(
                'UploaderTabsAll', 'Would you like to convert it to JPEG?')
            dialog.addButton(dialog.StandardButton.No)
        else:
            text += ' ' + translate(
                'UploaderTabsAll', 'Would you like to upload it anyway?')
        dialog.setInformativeText(text)
        dialog.setIcon(dialog.Icon.Warning)
        result = execute(dialog)
        if result == dialog.StandardButton.Yes:
            return ({}, {'convert': True})[convertible]
        if result == dialog.StandardButton.No:
            return {}
        return {'omit': True}

    def get_conversion_function(self, image, state, params):
        convert = {
            'omit': False,
            'resize': False,
            'convert': False,
            }
        mime_type = image.file_type
        if mime_type.startswith('video'):
            dims = image.metadata.dimensions
            if dims and dims['width'] and dims['height']:
                convert.update(self.ask_resize_image(
                    image, state, pixels=dims['width'] * dims['height']))
            if not any(convert.values()):
                convert.update(self.ask_resize_image(image, state))
            if not (any(convert.values()) or self.accepted_image_type(mime_type)):
                convert.update(self.ask_convert_image(
                    image, mime_type=mime_type, convertible=False))
            if any(convert.values()):
                return 'omit'
            return None
        readable = True
        try:
            tmp = self.data_to_image(self.read_image(image))
        except Exception as ex:
            logger.error(str(ex))
            readable = False
        if readable:
            convert.update(self.ask_resize_image(
                image, state, resizable=True,
                pixels=tmp['width'] * tmp['height']))
            mime_type = tmp['mime_type']
        if not (any(convert.values()) or self.accepted_image_type(mime_type)):
            convert.update(self.ask_convert_image(
                image, mime_type=mime_type, convertible=readable))
        if not any(convert.values()):
            convert.update(self.ask_resize_image(
                image, state, resizable=readable))
        if convert['omit']:
            return 'omit'
        if convert['resize'] or convert['convert']:
            return self.process_image
        if image.metadata.find_sidecar():
            # need to create file without sidecar
            return self.copy_file_and_metadata
        return None

    @QtSlot()
    @catch_all
    def stop_upload(self):
        if self.upload_worker:
            self.upload_worker.thread().requestInterruption()

    def get_selected_images(self):
        return self.app.image_list.get_selected_images()

    @QtSlot()
    @catch_all
    def start_upload(self):
        if not self.app.image_list.unsaved_files_dialog(with_discard=False):
            return
        # make list of items to upload
        upload_list = []
        state = {}
        for image in self.get_selected_images():
            params = self.get_upload_params(image, state)
            if params == 'abort':
                upload_list = []
                break
            if not params:
                continue
            convert = self.get_conversion_function(image, state, params)
            if convert == 'omit':
                continue
            upload_list.append((image, convert, params))
        if not upload_list:
            self.buttons['upload'].setChecked(False)
            return
        self.buttons['upload'].set_checked(True)
        self.enable_config(False)
        self.user_widget.connect_button.setEnabled(False)
        self.upload_progress({'busy': True})
        # do uploading in separate thread, so GUI can continue
        self.upload_worker = UploadWorker(self.user_widget.session, upload_list)
        thread = QtCore.QThread(self)
        self.upload_worker.moveToThread(thread)
        self.upload_worker.upload_progress.connect(self.upload_progress)
        thread.started.connect(self.upload_worker.start)
        self.upload_worker.finished.connect(self.uploader_finished)
        self.upload_worker.finished.connect(thread.quit)
        thread.finished.connect(thread.deleteLater)
        thread.start()

    @QtSlot(dict)
    @catch_all
    def upload_progress(self, update):
        if 'keyword' in update:
            # store photo id in image keywords, in main thread
            image, keyword = update['keyword']
            keywords = list(image.metadata.keywords)
            if keyword not in keywords:
                image.metadata.keywords = keywords + [keyword]
        if 'busy' in update:
            if update['busy']:
                self.progress_bar.setMaximum(0)
            else:
                self.progress_bar.setMaximum(100)
        if 'value' in update:
            self.progress_bar.setValue(update['value'])
        if 'label' in update:
            if update['label']:
                elided = self.progress_label.fontMetrics().elidedText(
                    update['label'], Qt.TextElideMode.ElideLeft,
                    width_for_text(self.progress_label, 'X' * 20))
                self.progress_label.setText(elided)
            else:
                self.progress_label.setText(
                    translate('UploaderTabsAll', 'Progress'))
        if 'error' in update:
            image, error = update['error']
            self.upload_error(os.path.basename(image.path), error)

    def upload_error(self, name, error):
        dialog = QtWidgets.QMessageBox(self)
        dialog.setWindowTitle(
            translate('UploaderTabsAll', 'Photini: upload error'))
        dialog.setText('<h3>{}</h3>'.format(translate(
            'UploaderTabsAll', 'File "{file_name}" upload failed.'
            ).format(file_name=name)))
        dialog.setInformativeText(error)
        dialog.setIcon(dialog.Icon.Warning)
        dialog.setStandardButtons(dialog.StandardButton.Abort |
                                  dialog.StandardButton.Ignore)
        dialog.setDefaultButton(dialog.StandardButton.Ignore)
        if execute(dialog) == dialog.StandardButton.Abort:
            self.stop_upload()

    @QtSlot()
    @catch_all
    def uploader_finished(self):
        self.buttons['upload'].set_checked(False)
        self.enable_config(True)
        self.user_widget.connect_button.setEnabled(True)
        self.upload_worker = None
        self.enable_upload_button()

    def new_selection(self, selection):
        self.enable_upload_button(selection=selection)

    def enable_upload_button(self, selection=None):
        if self.buttons['upload'].is_checked():
            # can always cancel upload in progress
            self.buttons['upload'].setEnabled(True)
            return
        if not self.user_widget.connect_button.is_checked():
            # can't upload if not logged in
            self.buttons['upload'].setEnabled(False)
            return
        if selection is None:
            selection = self.app.image_list.get_selected_images()
        self.buttons['upload'].setEnabled(len(selection) > 0)
        if 'sync' in self.buttons:
            self.buttons['sync'].setEnabled(
                len(selection) > 0 and self.user_widget.connect_button.is_checked())

    def get_upload_params(self, image, first_time=True):
        # only used by Flickr & Ipernity, which have a lot in common
        # get user preferences for this upload
        upload_prefs, replace_prefs, photo_id = self.replace_dialog(image)
        if not upload_prefs:
            # user cancelled dialog or chose to do nothing
            return None
        # set service dependent params
        params = self.get_params(image, upload_prefs, replace_prefs, photo_id)
        # add common metadata
        if upload_prefs['new_photo'] or replace_prefs['metadata']:
            lang = self.user_widget.user_data['lang']
            # title & description
            params['metadata'] = {}
            if image.metadata.title:
                params['metadata']['title'] = image.metadata.title.best_match(lang)
            else:
                params['metadata']['title'] = image.name
            description = []
            if image.metadata.headline:
                description.append(image.metadata.headline)
            if image.metadata.description:
                description.append(image.metadata.description.best_match(lang))
            if description:
                params['metadata']['description'] = '\n\n'.join(description)
            # keywords
            keywords = []
            if image.metadata.keywords:
                keywords = image.metadata.keywords.human_tags()
                for keyword, (ns, predicate,
                              value) in image.metadata.keywords.machine_tags():
                    if (ns != self.user_widget.client_data['name'] or
                            predicate not in ('photo_id', 'doc_id', 'id')):
                        keywords.append(keyword)
                keywords = [x.replace('"', "'") for x in keywords]
                for n, keyword in enumerate(keywords):
                    if ',' in keyword:
                        keywords[n] = '"' + keyword + '"'
            keywords.append('uploaded:by=photini')
            params['keywords'] = {'keywords': ','.join(keywords)}
        return params

    def replace_dialog(self, image, options, replace=True):
        # has image already been uploaded?
        photo_id = self.uploaded_id(image)
        if not photo_id:
            # new upload
            return {'new_photo': True}, {}, None
        # get user preferences
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle(translate('UploaderTabsAll', 'Replace photo'))
        dialog.setLayout(QtWidgets.QVBoxLayout())
        message = QtWidgets.QLabel(translate(
            'UploaderTabsAll', 'File {file_name} has already been uploaded to'
            ' {service}. How would you like to update it?').format(
                file_name=os.path.basename(image.path),
                service=self.user_widget.service_name()))
        message.setWordWrap(True)
        dialog.layout().addWidget(message)
        replace_options = {}
        for key, label in options:
            replace_options[key] = QtWidgets.QCheckBox(label)
        for key in self.replace_prefs:
            replace_options[key].setChecked(self.replace_prefs[key])
        upload_options = {}
        if replace:
            upload_options['replace_image'] = QtWidgets.QRadioButton(
                translate('UploaderTabsAll', 'Replace image'))
        upload_options['new_photo'] = QtWidgets.QRadioButton(
            translate('UploaderTabsAll', 'Upload as new photo'))
        upload_options['no_upload'] = QtWidgets.QRadioButton(
            translate('UploaderTabsAll', 'No image upload'))
        for key in self.upload_prefs:
            if self.upload_prefs[key]:
                upload_options[key].setChecked(True)
                break
        else:
            upload_options['no_upload'].setChecked(True)
        two_columns = QtWidgets.QHBoxLayout()
        column = QtWidgets.QVBoxLayout()
        for key in replace_options:
            upload_options['new_photo'].toggled.connect(
                replace_options[key].setDisabled)
            column.addWidget(replace_options[key])
        two_columns.addLayout(column)
        column = QtWidgets.QVBoxLayout()
        for key in upload_options:
            column.addWidget(upload_options[key])
        column.addStretch(1)
        two_columns.addLayout(column)
        dialog.layout().addLayout(two_columns)
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)
        dialog.layout().addWidget(button_box)
        if execute(dialog) != QtWidgets.QDialog.DialogCode.Accepted:
            return {}, {}, photo_id
        for key in replace_options:
            self.replace_prefs[key] = replace_options[key].isChecked()
        for key in upload_options:
            self.upload_prefs[key] = upload_options[key].isChecked()
        if self.upload_prefs['no_upload'] and not any(
                self.replace_prefs.values()):
            # user chose to do nothing
            return {}, {}, photo_id
        return dict(self.upload_prefs), dict(self.replace_prefs), photo_id

    def date_range(self, image):
        precision = min(image.metadata.date_taken['precision'], 6)
        min_taken_date = image.metadata.date_taken.truncate_datetime(
            image.metadata.date_taken['datetime'], precision)
        if precision >= 6:
            max_taken_date = min_taken_date + timedelta(seconds=1)
        elif precision >= 5:
            max_taken_date = min_taken_date + timedelta(minutes=1)
        elif precision >= 4:
            max_taken_date = min_taken_date + timedelta(hours=1)
        elif precision >= 3:
            max_taken_date = min_taken_date + timedelta(days=1)
        elif precision >= 2:
            max_taken_date = min_taken_date + timedelta(days=31)
        else:
            max_taken_date = min_taken_date + timedelta(days=366)
        max_taken_date -= timedelta(seconds=1)
        return min_taken_date, max_taken_date

    def find_local(self, unknowns, date_taken, icon_url):
        candidates = []
        for candidate in unknowns:
            if not candidate.metadata.date_taken:
                continue
            min_taken_date, max_taken_date = self.date_range(candidate)
            if date_taken < min_taken_date or date_taken > max_taken_date:
                continue
            candidates.append(candidate)
        if not candidates:
            return None
        # get user to choose matching image file
        rsp = requests.get(icon_url)
        if rsp.status_code == 200:
            remote_icon = rsp.content
        else:
            self.logger.error('HTTP error %d (%s)', rsp.status_code, icon_url)
            return None
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle(translate('UploaderTabsAll', 'Select an image'))
        dialog.setLayout(FormLayout())
        pixmap = QtGui.QPixmap()
        pixmap.loadFromData(remote_icon)
        label = QtWidgets.QLabel()
        label.setPixmap(pixmap)
        dialog.layout().addRow(label, Label(translate(
            'UploaderTabsAll',
            'Which image file matches this picture on {service}?').format(
                service=self.user_widget.service_name()), lines=2))
        buttons = {}
        frame = QtWidgets.QFrame()
        frame.setLayout(FormLayout())
        for candidate in candidates:
            label = QtWidgets.QLabel()
            pixmap = candidate.image.pixmap()
            if pixmap:
                label.setPixmap(pixmap)
            else:
                label.setText(candidate.image.text())
            button = QtWidgets.QPushButton(
                os.path.basename(candidate.path))
            button.setToolTip('<p>' + candidate.path + '</p>')
            button.setCheckable(True)
            button.clicked.connect(dialog.accept)
            frame.layout().addRow(label, button)
            buttons[button] = candidate
        button = QtWidgets.QPushButton(translate('UploaderTabsAll', 'No match'))
        button.setDefault(True)
        button.clicked.connect(dialog.reject)
        frame.layout().addRow('', button)
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidget(frame)
        dialog.layout().addRow(scroll_area)
        with UnBusy():
            if execute(dialog) == QtWidgets.QDialog.DialogCode.Accepted:
                for button, candidate in buttons.items():
                    if button.isChecked():
                        return candidate
        return None

    def uploaded_id(self, image):
        if not image.metadata.keywords:
            return None
        for keyword, (ns, predicate,
                      value) in image.metadata.keywords.machine_tags():
            if ns == self.user_widget.client_data['name'] and predicate in (
                    'photo_id', 'doc_id', 'id'):
                return value
        return None

    @QtSlot()
    @catch_all
    def sync_metadata(self):
        with Busy():
            # make list of known photo ids
            photo_ids = {}
            unknowns = []
            for image in self.app.image_list.get_selected_images():
                photo_id = self.uploaded_id(image)
                if photo_id:
                    photo_ids[photo_id] = image
                else:
                    unknowns.append(image)
            with self.user_widget.session(parent=self) as session:
                if unknowns:
                    # get date range of photos without an id
                    search_min, search_max = datetime.max, datetime.min
                    for image in unknowns:
                        if not image.metadata.date_taken:
                            continue
                        min_taken_date, max_taken_date = self.date_range(image)
                        search_min = min(search_min, min_taken_date)
                        search_max = max(search_max, max_taken_date)
                    # search remote service
                    for photo_id, date_taken, icon_url in session.find_photos(
                            search_min, search_max):
                        if photo_id in photo_ids:
                            continue
                        # find local image that matches remote date & icon
                        match = self.find_local(unknowns, date_taken, icon_url)
                        if match:
                            match.metadata.keywords = list(
                                match.metadata.keywords) + [
                                    '{}:id={}'.format(
                                        self.user_widget.client_data['name'],
                                        photo_id)]
                            photo_ids[photo_id] = match
                            unknowns.remove(match)
                # merge remote metadata into file
                for photo_id, image in photo_ids.items():
                    self.merge_metadata(session, photo_id, image)

    def merge_metadata_items(self, image, data):
        md = image.metadata
        for key, value in data.items():
            if key == 'keywords':
                value = [x for x in value if x != 'uploaded:by=photini']
            if value:
                old_value = getattr(md, key)
                if old_value:
                    value = old_value.merge(
                        '{}({})'.format(image.name, key),
                        self.user_widget.service_name(), value)
                setattr(md, key, value)

    def new_album_dialog(self):
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle(translate('UploaderTabsAll', 'Create new album'))
        dialog.setLayout(FormLayout())
        return dialog

    def exec_album_dialog(self, dialog):
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)
        dialog.layout().addRow(button_box)
        return execute(dialog) == QtWidgets.QDialog.DialogCode.Accepted



#  Photini - a simple photo metadata editor.
#  http://github.com/jim-easterbrook/Photini
#  Copyright (C) 2024  Jim Easterbrook  jim@jim-easterbrook.me.uk
#
#  This file is part of Photini.
#
#  Photini is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation, either version 3 of the License, or (at
#  your option) any later version.
#
#  Photini is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Photini.  If not, see <http://www.gnu.org/licenses/>.

import base64
import locale
import logging
import os

import requests

from photini.configstore import key_store
from photini.photinimap import GeocoderBase, PhotiniMap
from photini.pyqt import Busy, catch_all, Qt, QtCore, QtWidgets, scale_font
from photini.widgets import Label

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class AzureGeocoder(GeocoderBase):
    interval = 50
    api_key = key_store.get('azuremap', 'api_key')

    def query(self, params, url):
        params['subscription-key'] = self.api_key
        with Busy():
            self.rate_limit()
            try:
                rsp = requests.get(url, params=params, timeout=5)
            except Exception as ex:
                logger.error(str(ex))
                return []
        rsp = rsp.json()
        if 'error' in rsp:
            logger.error('Search error: %s', rsp['error']['message'])
            return []
        features = rsp['features']
        if not features:
            logger.error('No results found')
            return []
        return features

    def search(self, search_string, bounds=None):
        # see https://learn.microsoft.com/en-us/rest/api/maps/search/get-geocoding
        params = {
            'api-version': '2023-06-01',
            'query': search_string,
            'top': '20',
            }
        if bounds:
            north, east, south, west = bounds
            params['bbox'] = '{:.4f},{:.4f},{:.4f},{:.4f}'.format(
                west, south, east, north)
        lang, encoding = locale.getlocale()
        if lang:
            lang, sep, country = lang.partition('_')
            if country:
                params['view'] = country
        for feature in self.cached_query(
                params, 'https://atlas.microsoft.com/geocode'):
            properties = feature['properties']
            if properties['confidence'] == 'Low':
                continue
            address = properties['address']
            name = address['formattedAddress']
            if 'countryRegion' in address:
                country = address['countryRegion']['name']
                if country not in name:
                    name = '{}, {}'.format(name, country)
            if 'type' in properties:
                if properties['type'] in (
                        'AdminDivision1', 'AdminDivision2'):
                    continue
                if properties['type'] not in (
                        'Address', 'CountryRegion', 'PointOfInterest',
                        'PopulatedPlace'):
                    name = '{} ({})'.format(name, properties['type'])
            if 'bbox' in feature:
                west, south, east, north = feature['bbox']
            else:
                east, north = feature['geometry']['coordinates'][:2]
                west, south = east, north
            yield north, east, south, west, name

    def search_terms(self):
        widget = Label(
            translate(
                'MapTabAzure', 'Search provided by Microsoft Azure'),
            lines=2)
        widget.setAlignment(Qt.AlignmentFlag.AlignRight)
        scale_font(widget, 80)
        return [widget]


class TabWidget(PhotiniMap):
    api_key = key_store.get('azuremap', 'api_key')

    @staticmethod
    def tab_name():
        return translate('MapTabAzure', 'Map (&Azure)')

    def get_geocoder(self):
        return AzureGeocoder(parent=self)

    def get_head(self):
        data_dir = os.path.join(os.path.dirname(__file__), 'data')
        with open(os.path.join(data_dir, 'map', 'circle_blue.png'), 'rb') as f:
            circle_blue_data = f.read()
        with open(os.path.join(data_dir, 'map', 'circle_red.png'), 'rb') as f:
            circle_red_data = f.read()
        url_base = 'https://atlas.microsoft.com/sdk/javascript/mapcontrol/3/'
        url_js = url_base + 'atlas.min.js'
        url_css = url_base + 'atlas.min.css'
        return '''<script type="text/javascript">
var circle_blue_data = "data:image/png;base64,{circle_blue_data}";
var circle_red_data = "data:image/png;base64,{circle_red_data}";
    </script>
    <link rel="stylesheet" href="{url_css}" type="text/css" />
    <script type="text/javascript" src="{url_js}"></script>
    <script type="text/javascript" src="azuremap.js"></script>'''.format(
        url_css=url_css, url_js=url_js,
        circle_blue_data=base64.b64encode(circle_blue_data).decode('ascii'),
        circle_red_data=base64.b64encode(circle_red_data).decode('ascii'))

    def get_body(self, text_dir):
        return '''  <body onload="initialize()" ondragstart="return false">
    <div id="mapDiv" dir="{text_dir}"></div>
  </body>'''.format(text_dir=text_dir)

    def get_options(self):
        options = {
            'authOptions': {
                'authType': 'subscriptionKey',
                'subscriptionKey': self.api_key,
                },
            }
        language, encoding = locale.getlocale()
        if language:
            options['language'] = language.replace('_', '-')
            language, sep, country = language.partition('_')
            if country:
                options['View'] = country
        return options



#  Photini - a simple photo metadata editor.
#  http://github.com/jim-easterbrook/Photini
#  Copyright (C) 2023  Jim Easterbrook  jim@jim-easterbrook.me.uk
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or (at
#  your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from setuptools import build_meta as _orig
from setuptools.build_meta import *


def build_lang():
    src_dir = os.path.join('src', 'lang')
    dst_dir = os.path.join('src', 'photini', 'data', 'lang')
    if os.path.exists(dst_dir) and not os.path.exists(src_dir):
        # probably building wheel from sdist, no need to compile
        return
    print('running build_lang')
    os.makedirs(dst_dir, exist_ok=True)
    for lang in os.listdir(src_dir):
        src_file = os.path.join(src_dir, lang, 'photini.ts')
        if not os.path.exists(src_file):
            continue
        dst_file = os.path.join(dst_dir, 'photini.' + lang + '.qm')
        if (os.path.exists(dst_file) and
                os.stat(dst_file).st_mtime >= os.stat(src_file).st_mtime):
            continue
        print('compiling {} -> {}'.format(src_file, dst_file))
        try:
            subprocess.check_call([
                'lrelease-qt5', '-silent', src_file, '-qm', dst_file])
        except FileNotFoundError as ex:
            print('FAIL:', str(ex))
            return


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    build_lang()
    return _orig.build_wheel(wheel_directory, config_settings=config_settings,
                             metadata_directory=metadata_directory)


def build_sdist(sdist_directory, config_settings=None):
    build_lang()
    return _orig.build_sdist(sdist_directory, config_settings=config_settings)



#  Photini - a simple photo metadata editor.
#  http://github.com/jim-easterbrook/Photini
#  Copyright (C) 2012-24  Jim Easterbrook  jim@jim-easterbrook.me.uk
#
#  This program is free software: you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see
#  <http://www.gnu.org/licenses/>.

from setuptools import setup
from setuptools import __version__ as setuptools_version

if tuple(map(int, setuptools_version.split('.')[:2])) >= (61, 0):
    # use metadata from pyproject.toml directly
    setup()
else:
    import os

    # get metadata from pyproject.toml
    import toml
    metadata = toml.load('pyproject.toml')

    with open(metadata['project']['readme']) as ldf:
        long_description = ldf.read()

    package_data = []
    for root, dirs, files in os.walk('src/photini/data/'):
        package_data += [
            os.path.join(root.replace('src/photini/', ''), x) for x in files]

    setup(name = metadata['project']['name'],
          author = metadata['project']['authors'][0]['name'],
          author_email = metadata['project']['authors'][0]['email'],
          url = metadata['project']['urls']['Homepage'],
          description = metadata['project']['description'],
          long_description = long_description,
          classifiers = metadata['project']['classifiers'],
          license = metadata['project']['license']['text'],
          packages = ['photini'],
          package_dir = {'' : 'src'},
          package_data = {'photini' : package_data},
          entry_points = {
              'console_scripts' : [
                  '{} = {}'.format(k, v)
                  for k, v in metadata['project']['scripts'].items()],
              'gui_scripts' : [
                  '{} = {}'.format(k, v)
                  for k, v in metadata['project']['gui-scripts'].items()],
              },
          install_requires = metadata['project']['dependencies'],
          extras_require = metadata['project']['optional-dependencies'],
          zip_safe = metadata['tool']['setuptools']['zip-safe'],
          )



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import codecs
from collections import defaultdict
import logging
import os
import re

import exiv2
import chardet

logger = logging.getLogger(__name__)

exiv2_version = 'python-exiv2 {}, exiv2 {}'.format(
    exiv2.__version__, exiv2.version())


class MetadataHandler(object):
    # static data about keys known to Exiv2
    data_sets = {}
    md_info = {}

    @classmethod
    def get_info(cls, tag_name):
        if tag_name in cls.md_info:
            return cls.md_info[tag_name]
        family, group, tag = tag_name.split('.')
        if family == 'Iptc':
            if group not in cls.data_sets:
                # get static info once, as it's computed each time
                if group == 'Application2':
                    cls.data_sets[group] = exiv2.IptcDataSets.application2RecordList()
                elif group == 'Envelope':
                    cls.data_sets[group] = exiv2.IptcDataSets.envelopeRecordList()
            for data_set in cls.data_sets[group]:
                # case insensitive as Iptc Sublocation is SubLocation in Exiv2
                if data_set['name'].lower() == tag.lower():
                    cls.md_info[tag_name] = data_set
                    break
        if tag_name not in cls.md_info:
            cls.md_info[tag_name] = {}
        return cls.md_info[tag_name]

    @classmethod
    def initialise(cls, config_store, verbosity):
        level = min(exiv2.LogMsg.Level.error, 4 - verbosity)
        level = max(exiv2.LogMsg.Level.debug, level)
        exiv2.LogMsg.setLevel(exiv2.LogMsg.Level(level))
        exiv2.XmpParser.initialize()
        if exiv2.__version_tuple__ < (0, 17) and exiv2.testVersion(0, 27, 4):
            exiv2.enableBMFF(True)
        if config_store:
            config_store.delete('metadata', 'enable_bmff')
        # Recent versions of Exiv2 have these namespaces defined, but
        # older versions may not recognise them. The xapGImg URL is
        # invalid, but Photini doesn't write xapGImg so it doesn't
        # matter.
        registered = exiv2.XmpProperties.registeredNamespaces()
        for prefix, ns in (
                ('exifEX',      'http://cipa.jp/exif/1.0/'),
                ('plus',        'http://ns.useplus.org/ldf/xmp/1.0/'),
                ('xmp',         'http://ns.adobe.com/xap/1.0/'),
                ('xapGImg',     'http://ns.adobe.com/xxx/'),
                ('xmpGImg',     'http://ns.adobe.com/xap/1.0/g/img/'),
                ('xmpRights',   'http://ns.adobe.com/xap/1.0/rights/'),
                ):
            if prefix not in registered:
                exiv2.XmpProperties.registerNs(ns, prefix)

    def __init__(self, path=None, buf=None):
        self._path = path
        # read metadata
        if buf:
            self._image = exiv2.ImageFactory.open(buf)
        else:
            self._image = exiv2.ImageFactory.open(self._path)
        if self._path:
            self._name = os.path.basename(self._path)
        else:
            self._name = 'data'
        self._image.readMetadata()
        self._exifData = self._image.exifData()
        self._iptcData = self._image.iptcData()
        self._xmpData = self._image.xmpData()
        access_mode = {
            'exif': self._image.checkMode(exiv2.MetadataId.Exif),
            'iptc': self._image.checkMode(exiv2.MetadataId.Iptc),
            'xmp': self._image.checkMode(exiv2.MetadataId.Xmp),
            }
        self.read_only = not any(
            [x in (exiv2.AccessMode.Write, exiv2.AccessMode.ReadWrite)
             for x in access_mode.values()])
        self.mime_type = self._image.mimeType()
        if self.mime_type == 'image/generic':
            self.mime_type = None
        self.xmp_only = self.mime_type in (
            'application/rdf+xml', 'application/postscript')
        # Don't use Exiv2's converted values when accessing Xmp files
        if self.xmp_only:
            self.clear_exif()
            self.clear_iptc()
        # remove old software tag
        tag = 'Exif.Image.ProcessingSoftware'
        if (tag in self._exifData
                and str(self._exifData[tag].value()).startswith('Photini')):
            del self._exifData[tag]
        # transcode any non utf-8 strings (Xmp is always utf-8)
        encodings = []
        iptc_charset = self.get_iptc_encoding()
        if iptc_charset not in ('utf-8', 'ascii', None):
            iptc_charset = codecs.lookup(iptc_charset).name
            logger.info('%s: IPTC character set %s', self._name, iptc_charset)
            encodings.append(iptc_charset)
        for data, set_value in ((self._exifData, self.set_exif_value),
                                (self._iptcData, self.set_iptc_value)):
            new_data = defaultdict(list)
            for datum in data:
                if datum.typeId() not in (exiv2.TypeId.asciiString,
                                          exiv2.TypeId.string):
                    continue
                key = datum.key()
                if key in ('Iptc.Envelope.CharacterSet', 'Exif.Image.IPTCNAA'):
                    continue
                if '.0x' in key:
                    # unknown key type
                    continue
                raw_value = datum.value().data()
                if self.decode_string(key, raw_value, 'utf-8') is not None:
                    # no need to do anything
                    continue
                for encoding in encodings:
                    value = self.decode_string(key, raw_value, encoding)
                    if value:
                        break
                else:
                    encoding = chardet.detect(bytearray(raw_value))['encoding']
                    if encoding:
                        try:
                            encoding = codecs.lookup(encoding).name
                        except LookupError:
                            encoding = None
                    if encoding:
                        logger.info("%s: detected character set '%s'",
                                    self._name, encoding)
                        value = self.decode_string(key, raw_value, encoding)
                    else:
                        value = None
                    if value:
                        encodings.append(encoding)
                    else:
                        logger.warning('%s: failed to transcode %s',
                                       self._name, key)
                        value = bytes(raw_value).decode(
                            'utf-8', errors='replace')
                new_data[key].append(value)
            for key, value in new_data.items():
                if len(value) == 1:
                    set_value(key, value[0])
                else:
                    set_value(key, value)
        # assume no XMP thumbnails to replace or append
        self._xmp_thumb_idx = None
        # mapping Xmp data to type_id, initialised with some that exiv2
        # doesn't know, extended with ones read from files
        self._xmp_type_id = {
            'Xmp.iptc.AltTextAccessibility': exiv2.TypeId.langAlt,
            'Xmp.iptc.ExtDescrAccessibility': exiv2.TypeId.langAlt,
            'Xmp.iptcExt.ImageRegion': exiv2.TypeId.xmpBag,
            'Iptc4xmpExt:LocationName': exiv2.TypeId.langAlt,
            'Iptc4xmpExt:Name': exiv2.TypeId.langAlt,
            'Iptc4xmpExt:rCtype': exiv2.TypeId.xmpBag,
            'Iptc4xmpExt:rRole': exiv2.TypeId.xmpBag,
            'Iptc4xmpExt:rbVertices': exiv2.TypeId.xmpSeq,
            'Xmp.xmp.Thumbnails': exiv2.TypeId.xmpAlt,
            }

    def set_xmp_type(self, key, value):
        if value == exiv2.TypeId.xmpText:
            return
        sub_key = key.split('/')[-1]
        if sub_key in self._xmp_type_id:
            if value == self._xmp_type_id[sub_key]:
                return
            logger.warning('altered type_id %s', sub_key)
        self._xmp_type_id[sub_key] = value

    def get_xmp_type(self, key):
        value = exiv2.XmpProperties.propertyType(exiv2.XmpKey(key))
        if value != exiv2.TypeId.xmpText:
            # exiv2 knows what it should be
            self.set_xmp_type(key, value)
            return value
        sub_key = key.split('/')[-1]
        if sub_key in self._xmp_type_id:
            return self._xmp_type_id[sub_key]
        return exiv2.TypeId.xmpText

    def decode_string(self, tag, raw_value, encoding):
        try:
            result = codecs.decode(raw_value, encoding=encoding)
        except UnicodeDecodeError:
            return None
        if encoding != 'utf-8':
            logger.info(
                "%s: transcoded %s from '%s'", self._name, tag, encoding)
        return result

    def get_iptc_encoding(self):
        # IPTC-IIM uses ISO/IEC 2022 escape sequences to set the
        # character set. These are very flexible, and it's not obvious
        # which sequences, if any, are ever used in practice. The
        # sequence is escape, one or more intermediate bytes, then a
        # final byte.
        if 'Iptc.Envelope.CharacterSet' not in self._iptcData:
            return None
        iptc_charset_code = self._iptcData[
            'Iptc.Envelope.CharacterSet'].value().data()
        if len(iptc_charset_code) >= 3 and iptc_charset_code[0] == 0x1b:
            intermediate = iptc_charset_code[1:-1]
            final = iptc_charset_code[-1]
            if intermediate == b'\x25':
                # "other coding systems"
                if final == 0x47:
                    return 'utf-8'
            if intermediate in (b'\x28', b'\x29', b'\x2a', b'\x2b'):
                # "94-character set"
                if final == 0x4e:
                    return 'koi8_r'
                return 'ascii'
            if intermediate in (b'\x2c', b'\x2d', b'\x2e', b'\x2f'):
                # "96-character set"
                if final == 0x40:
                    return 'iso_8859_5'
                if final == 0x41:
                    return 'iso_8859_1'
                if final == 0x42:
                    return 'iso_8859_2'
                if final == 0x43:
                    return 'iso_8859_3'
                if final == 0x44:
                    return 'iso_8859_4'
                if final == 0x46:
                    return 'iso_8859_7'
                if final == 0x47:
                    return 'iso_8859_6'
                if final == 0x48:
                    return 'iso_8859_8'
            # "multi-byte character set"
            if intermediate in (b'\x24\x28', b'\x24\x29',
                                b'\x24\x2a', b'\x24\x2b'):
                if final == 0x42:
                    return 'euc_jp'
            if intermediate == b'\x25\x2f':
                if final == 0x46:
                    return 'utf-32-be'
                if final == 0x4c:
                    return 'utf-16-be'
        logger.error('Unrecognised IPTC character set %s',
                     repr(bytes(iptc_charset_code)))
        return None

    def set_iptc_encoding(self):
        # escape % G, or \x1b \x25 \x47, selects UTF-8
        self.set_iptc_value('Iptc.Envelope.CharacterSet', '\x1b%G')

    def clear_exif(self):
        self._exifData.clear()
        self._image.clearExifData()

    def clear_iptc(self):
        self._iptcData.clear()
        self._image.clearIptcData()

    def get_exif_tags(self):
        for datum in self._exifData:
            yield datum.key()

    def get_iptc_tags(self):
        for datum in self._iptcData:
            yield datum.key()

    def get_xmp_tags(self):
        for datum in self._xmpData:
            yield datum.key()

    @classmethod
    def open_old(cls, path, *arg, quiet=False, **kw):
        try:
            return cls(path, *arg, **kw)
        except exiv2.Exiv2Error as ex:
            # expected if unrecognised file format
            if quiet:
                logger.info(str(ex))
            else:
                logger.warning(str(ex))
            return None
        except Exception as ex:
            logger.error('Exception opening %s', path)
            logger.exception(ex)
            return None

    def set_exif_thumbnail_from_buffer(self, buffer):
        thumb = exiv2.ExifThumb(self._exifData)
        thumb.setJpegThumbnail(buffer)

    def get_exif_comment(self, tag, value):
        if isinstance(value, exiv2.CommentValue):
            data = value.data()
            charset = value.charsetId()
        else:
            data = bytearray(len(value))
            value.copy(data, exiv2.ByteOrder.invalidByteOrder)
            if data[:5] == b'ASCII':
                charset = exiv2.CommentValue.CharsetId.ascii
            elif data[:3] == b'JIS':
                charset = exiv2.CommentValue.CharsetId.jis
            elif data[:7] == b'UNICODE':
                charset = exiv2.CommentValue.CharsetId.unicode
            else:
                charset = exiv2.CommentValue.CharsetId.undefined
        # ignore Exiv2's comment decoding, Python is better at unicode
        if not any(data):
            return None
        raw_value = data[8:]
        if charset == exiv2.CommentValue.CharsetId.ascii:
            encodings = ('ascii',)
        elif charset == exiv2.CommentValue.CharsetId.jis:
            # Exif standard says JIS X208-1990, but doesn't say what encoding
            encodings = ('iso2022_jp', 'euc_jp', 'shift_jis', 'cp932')
        elif charset == exiv2.CommentValue.CharsetId.unicode:
            # Exif refers to 1991 unicode, which predates UTF-8.
            # Check for BOM.
            if raw_value[:3] == b'\xef\xbb\xbf':
                raw_value = raw_value[3:]
                encodings = ('utf_8',)
            elif raw_value[:2] == b'\xff\xfe':
                raw_value = raw_value[2:]
                encodings = ('utf_16_le',)
            elif raw_value[:2] == b'\xfe\xff':
                raw_value = raw_value[2:]
                encodings = ('utf_16_be',)
            else:
                # If no BOM, utf-16 should be bigendian so try that first.
                encodings = ('utf_16_be', 'utf_16_le', 'utf_8')
        else:
            encodings = ()
            if charset != exiv2.CommentValue.CharsetId.undefined:
                logger.warning('%s: %s: unknown charset', self._name, tag)
                raw_value = data
        result = None
        for encoding in encodings:
            result = self.decode_string(tag, raw_value, encoding)
            if result:
                break
        else:
            detector = chardet.universaldetector.UniversalDetector()
            for i in range(0, len(raw_value), 100):
                detector.feed(raw_value[i:i+100])
                if detector.done:
                    break
            detector.close()
            encoding = detector.result['encoding']
            if encoding:
                result = self.decode_string(tag, raw_value, encoding)
        if result and '\0' in result:
            # terminating NULLs are allowed
            result = result.strip('\0')
            if '\0' in result:
                # NULLs within the string are not allowed
                result = None
        if not result:
            logger.error(
                '%s: %s: %d bytes binary data will be deleted when metadata'
                ' is saved', self._name, tag, value.size())
            return None
        return result

    def get_exif_value(self, tag):
        try:
            key = exiv2.ExifKey(tag)
        except exiv2.Exiv2Error:
            # old versions of libexiv2 don't recognise newer tags
            return None
        datum = self._exifData.findKey(key)
        if datum == self._exifData.end():
            return None
        if tag in ('Exif.Canon.ModelID', 'Exif.CanonCs.LensType',
                   'Exif.Canon.SerialNumber', 'Exif.CanonLe.LensSerialNumber',
                   'Exif.Image.XPTitle', 'Exif.Image.XPComment',
                   'Exif.Image.XPAuthor', 'Exif.Image.XPKeywords',
                   'Exif.Image.XPSubject', 'Exif.NikonLd1.LensIDNumber',
                   'Exif.Minolta.LensID', 'Exif.Nikon3.LensType',
                   'Exif.NikonLd2.LensIDNumber', 'Exif.NikonLd3.LensIDNumber',
                   'Exif.NikonLd4.LensIDNumber', 'Exif.OlympusEq.LensType',
                   'Exif.Olympus2.CameraID',
                   'Exif.Panasonic.InternalSerialNumber',
                   'Exif.Pentax.LensType', 'Exif.Pentax.ModelID',
                   'Exif.PentaxDng.LensType', 'Exif.PentaxDng.ModelID',
                   'Exif.Sony1.LensID', 'Exif.Sony1.SonyModelID',
                   'Exif.Sony2.LensID', 'Exif.Sony2.SonyModelID'):
            # use Exiv2's "interpreted string"
            if exiv2.__version_tuple__ >= (0, 16, 2):
                return datum.print(self._exifData)
            else:
                return datum._print(self._exifData)
        value = datum.value()
        if tag in ('Exif.Photo.UserComment',
                   'Exif.GPSInfo.GPSProcessingMethod'):
            return self.get_exif_comment(tag, value)
        if isinstance(value, exiv2.AsciiValue):
            return value.toString()
        if isinstance(value, exiv2.DataValue):
            result = bytearray(value.size())
            value.copy(result, exiv2.ByteOrder.invalidByteOrder)
            return result
        if isinstance(value, (exiv2.RationalValue, exiv2.URationalValue,
                              exiv2.ShortValue, exiv2.UShortValue,
                              exiv2.LongValue, exiv2.ULongValue)):
            if len(value) > 1:
                return list(value)
            if len(value) == 0:
                return None
            return value[0]
        logger.warning(
            '%s: %s: reading %s as string', self._name, tag, type(value))
        return value.toString()

    def decode_iptc_value(self, datum):
        type_id = datum.typeId()
        value = datum.value()
        if type_id == exiv2.TypeId.date:
            return value.getDate()
        if type_id == exiv2.TypeId.time:
            return value.getTime()
        return value.toString()

    def get_iptc_value(self, tag):
        result = None
        # findKey gets first matching datum and returns an iterator
        # which we use to search the rest of the data
        for datum in self._iptcData.findKey(exiv2.IptcKey(tag)):
            if result is None:
                # first datum
                result = self.decode_iptc_value(datum)
                if not exiv2.IptcDataSets.dataSetRepeatable(
                                        datum.tag(), datum.record()):
                    break
                result = [result]
            elif datum.key() == tag:
                result.append(self.decode_iptc_value(datum))
        return result

    _re_key_parts = re.compile(r'(.*?)(\[(\d+)\])?(/(.*))?$')

    def set_item(self, result, key, value):
        # break exiv2 tags into component parts
        match = self._re_key_parts.match(key)
        root = match.group(1)
        idx = match.group(3)
        node = match.group(5)
        if node:
            if idx:
                self.set_item(result[root][int(idx)-1], node, value)
            else:
                self.set_item(result[root], node, value)
        elif idx:
            result[root].append(value)
        else:
            result[root] = value

    def get_xmp_value(self, tag):
        # XMP has a nested structure of arbitrary depth. Exiv2 converts
        # this to "flat" tag names. This method converts back to nested
        # dicts and lists.
        result = {}
        for datum in self._xmpData:
            key = datum.key()
            if not key.startswith(tag):
                continue
            value = datum.value()
            type_id = value.typeId()
            if type_id == exiv2.TypeId.langAlt:
                value = dict(value)
            elif type_id in (exiv2.TypeId.xmpAlt, exiv2.TypeId.xmpBag,
                             exiv2.TypeId.xmpSeq):
                value = list(value)
            elif tag == 'Xmp.xmp.Thumbnails' and ':image' in key:
                # don't copy large string to unicode
                value = value.data()
            elif value.xmpStruct() == exiv2.XmpValue.XmpStruct.xsStruct:
                value = {}
            else:
                array_type = value.xmpArrayType()
                if array_type == exiv2.XmpValue.XmpArrayType.xaNone:
                    value = str(value)
                else:
                    value = []
                    type_id = {
                        exiv2.XmpValue.XmpArrayType.xaAlt: exiv2.TypeId.xmpAlt,
                        exiv2.XmpValue.XmpArrayType.xaBag: exiv2.TypeId.xmpBag,
                        exiv2.XmpValue.XmpArrayType.xaSeq: exiv2.TypeId.xmpSeq,
                        }[array_type]
            self.set_xmp_type(key, type_id)
            self.set_item(result, key, value)
        if tag in result:
            return result[tag]
        return None

    def select_exif_thumbnail(self):
        if self.xmp_only:
            return
        # try normal thumbnail
        thumb = exiv2.ExifThumb(self._exifData)
        data = thumb.copy()
        if data:
            logger.info('%s: trying thumbnail', self._name)
            yield data, 'thumbnail'
        # try preview images
        preview_manager = exiv2.PreviewManager(self._image)
        props = preview_manager.getPreviewProperties()
        if not props:
            return
        # get largest acceptable images
        idx = len(props)
        while idx > 0:
            idx -= 1
            if max(props[idx].width_, props[idx].height_) <= 640:
                logger.info('%s: trying preview %dx%d', self._name,
                            props[idx].width_, props[idx].height_)
                image = preview_manager.getPreviewImage(props[idx])
                yield image, 'preview ' + str(idx)

    def select_xmp_thumbnail(self, file_value):
        if not file_value:
            return
        # new thumbnail will be appended
        self._xmp_thumb_idx = len(file_value) + 1
        # make list of all thumbnails
        candidates = []
        for n, value in enumerate(file_value):
            candidate = dict((k.split(':')[1], v) for (k, v) in value.items())
            candidate['width'] = int(candidate['width'])
            candidate['height'] = int(candidate['height'])
            if max(candidate['width'], candidate['height']) == 160:
                # new thumbnail will replace this one
                self._xmp_thumb_idx = n + 1
                # try this one before any others
                logger.info('%s: trying xmp thumb %dx%d', self._name,
                            candidate['width'], candidate['height'])
                yield candidate['image'], 'xmp thumb ' + str(n)
            else:
                candidates.append(candidate)
        # Xmp spec says first thumbnail is default, so just try them in order
        for n, candidate in enumerate(candidates):
            logger.info('%s: trying xmp thumb %dx%d', self._name,
                        candidate['width'], candidate['height'])
            yield candidate['image'], 'xmp thumb ' + str(n)

    def get_previews(self):
        preview_manager = exiv2.PreviewManager(self._image)
        props = preview_manager.getPreviewProperties()
        idx = len(props)
        while idx:
            idx -= 1
            yield preview_manager.getPreviewImage(props[idx])

    def get_preview_imagedims(self):
        preview_manager = exiv2.PreviewManager(self._image)
        props = preview_manager.getPreviewProperties()
        for p in props:
            yield p.width_, p.height_

    def set_exif_value(self, tag, value):
        if not value:
            self.clear_exif_tag(tag)
            return
        key = exiv2.ExifKey(tag)
        type_id = key.defaultTypeId()
        if type_id == exiv2.TypeId.unsignedByte:
            value = exiv2.DataValue(
                value, exiv2.ByteOrder.invalidByteOrder, type_id)
        elif type_id == exiv2.TypeId.asciiString:
            value = exiv2.AsciiValue(value)
        elif type_id == exiv2.TypeId.comment:
            # only comment value Photini writes is GPS processing method
            # which is certain to be ASCII
            value = 'charset=Ascii ' + value
            value = exiv2.CommentValue(value)
        elif type_id == exiv2.TypeId.unsignedShort:
            value = exiv2.UShortValue(value)
        elif type_id == exiv2.TypeId.unsignedLong:
            value = exiv2.ULongValue(value)
        elif type_id == exiv2.TypeId.unsignedRational:
            if isinstance(value, (list, tuple)):
                value = exiv2.URationalValue(
                    [(x.numerator, x.denominator) for x in value])
            else:
                value = exiv2.URationalValue(
                    [(value.numerator, value.denominator)])
        else:
            # unhandled type, use the string representation
            logger.warning('%s: %s: writing %s type as string',
                           self._name, tag, exiv2.TypeInfo.typeName(type_id))
        datum = self._exifData[tag]
        datum.setValue(value)

    @classmethod
    def iptc_max_len(cls, tag_name):
        data_set = cls.get_info(tag_name)
        if 'maxbytes' in data_set:
            return data_set['maxbytes']
        return None

    @classmethod
    def max_bytes(cls, name):
        result = []
        for mode, tag in cls._tag_list[name]:
            if mode == 'WA' and tag.split('.')[0] == 'Iptc':
                if tag in cls._multi_tags:
                    for sub_tag in cls._multi_tags[tag]:
                        if sub_tag:
                            result.append(cls.iptc_max_len(sub_tag))
                else:
                    result.append(cls.iptc_max_len(tag))
        result = [x for x in result if x]
        if result:
            return min(result)
        return None

    @classmethod
    def truncate_iptc(cls, tag, value):
        max_bytes = cls.iptc_max_len(tag)
        if max_bytes:
            value = value.encode('utf-8')[:max_bytes]
            value = value.decode('utf-8', errors='ignore')
        return value

    def set_iptc_value(self, tag, value):
        if not value:
            self.clear_iptc_tag(tag)
            return
        # make list of values
        key = exiv2.IptcKey(tag)
        type_id = exiv2.IptcDataSets.dataSetType(key.tag(), key.record())
        if type_id == exiv2.TypeId.date:
            values = [exiv2.DateValue(*value)]
        elif type_id == exiv2.TypeId.time:
            values = [exiv2.TimeValue(*value)]
        elif isinstance(value, (list, tuple)):
            values = [self.truncate_iptc(tag, x) for x in value]
        else:
            values = [self.truncate_iptc(tag, value)]
        # update or delete existing values
        datum = self._iptcData.findKey(key)
        while datum != self._iptcData.end():
            if datum.key() == tag:
                if values:
                    datum.setValue(values.pop(0))
                else:
                    datum = self._iptcData.erase(datum)
                    continue
            next(datum)
        # append remaining values
        while values:
            datum = exiv2.Iptcdatum(key)
            datum.setValue(values.pop(0))
            if self._iptcData.add(datum) != 0:
                logger.error('%s: duplicated tag %s', self._name, tag)
                return

    def set_xmp_value(self, tag, value):
        if not value:
            self.clear_xmp_tag(tag)
            return
        type_id = self.get_xmp_type(tag)
        if type_id == exiv2.TypeId.langAlt:
            self._xmpData[tag] = exiv2.LangAltValue(value)
        elif isinstance(value, dict):
            # clear any existing struct members
            self.clear_xmp_tag(tag, children_only=True)
            # create struct
            tmp = exiv2.XmpTextValue()
            tmp.setXmpStruct()
            self._xmpData[tag] = tmp
            # set struct members
            for k, v in value.items():
                self.set_xmp_value('{}/{}'.format(tag, k), v)
        elif not isinstance(value, (list, tuple)):
            # simple value
            self._xmpData[tag] = exiv2.XmpTextValue(str(value))
        elif not isinstance(value[0], dict):
            # simple array value
            self._xmpData[tag] = exiv2.XmpArrayValue(value, type_id)
        else:
            # clear any existing array elements
            self.clear_xmp_tag(tag, children_only=True)
            # create array
            array_type = {
                exiv2.TypeId.xmpAlt: exiv2.XmpValue.XmpArrayType.xaAlt,
                exiv2.TypeId.xmpBag: exiv2.XmpValue.XmpArrayType.xaBag,
                exiv2.TypeId.xmpSeq: exiv2.XmpValue.XmpArrayType.xaSeq,
                }[type_id]
            tmp = exiv2.XmpTextValue()
            tmp.setXmpArrayType(array_type)
            self._xmpData[tag] = tmp
            # set array elements
            for idx, element in enumerate(value):
                self.set_xmp_value('{}[{}]'.format(tag, idx+1), element)

    def clear_exif_tag(self, tag):
        datum = self._exifData.findKey(exiv2.ExifKey(tag))
        if datum != self._exifData.end():
            self._exifData.erase(datum)

    def clear_iptc_tag(self, tag):
        # findKey gets first matching datum and returns an iterator
        # which we use to search the rest of the data
        datum = self._iptcData.findKey(exiv2.IptcKey(tag))
        while datum != self._iptcData.end():
            if datum.key() == tag:
                datum = self._iptcData.erase(datum)
            else:
                next(datum)

    def clear_xmp_tag(self, tag, children_only=False):
        key = exiv2.XmpKey(tag)
        datum = self._xmpData.findKey(key)
        if datum == self._xmpData.end():
            return datum
        value = datum.value()
        type_id = value.typeId()
        if type_id == exiv2.TypeId.xmpText:
            # can be a struct or array
            if value.xmpStruct() == exiv2.XmpValue.XmpStruct.xsStruct:
                self.clear_xmp_struct(tag)
            elif value.xmpArrayType() != exiv2.XmpValue.XmpArrayType.xaNone:
                self.clear_xmp_array(tag)
        if children_only:
            return datum
        return self._xmpData.erase(datum)

    def clear_xmp_struct(self, tag):
        tag += '/'
        datum = self._xmpData.begin()
        while datum != self._xmpData.end():
            key = datum.key()
            if key.startswith(tag):
                datum = self.clear_xmp_tag(key)
            else:
                next(datum)

    def clear_xmp_array(self, tag):
        idx = 0
        while True:
            idx += 1
            key = exiv2.XmpKey('{}[{}]'.format(tag, idx))
            if self._xmpData.findKey(key) == self._xmpData.end():
                return
            while self.clear_xmp_tag(str(key)) != self._xmpData.end():
                pass

    def has_iptc(self):
        return self._iptcData.count() > 0

    def has_exif_tag(self, tag):
        return tag in self._exifData

    def save_file(self):
        try:
            self._image.writeMetadata()
        except exiv2.Exiv2Error as ex:
            logger.error(str(ex))
            return False
        except Exception as ex:
            logger.exception(ex)
            return False
        return True

    def clear_gps(self):
        for data in self._exifData, self._xmpData:
            pos = data.begin()
            while pos != data.end():
                tag = pos.key().split('.')[2]
                if tag.startswith('GPS'):
                    pos = data.erase(pos)
                else:
                    next(pos)

    def clear_maker_note(self):
        self.clear_exif_tag('Exif.Image.Make')
        self._image.writeMetadata()
        self._image.readMetadata()
        self._exifData = self._image.exifData()
        self.clear_exif_tag('Exif.Photo.MakerNote')

    @staticmethod
    def create_sc(path, image_md):
        image = exiv2.ImageFactory.create(exiv2.ImageType.xmp, path)
        image.writeMetadata()
        if image_md:
            # let exiv2 copy as much metadata as it can into sidecar
            image.setMetadata(image_md._image)
            image.writeMetadata()

    def merge_sc(self, other):
        # open other image and read its metadata
        image = exiv2.ImageFactory.open(other._path)
        image.readMetadata()
        # copy Exif data inferred by libexiv2
        for o_datum in image.exifData():
            tag = o_datum.key()
            s_datum = self._exifData[tag]
            s_datum.setValue(o_datum.value())
        # copy Xmp data, except inferred Exif data
        for o_datum in image.xmpData():
            tag = o_datum.key()
            if tag.startswith('Xmp.xmp.Thumbnails'):
                continue
            ns = tag.split('.')[1]
            if ns in ('exif', 'exifEX', 'tiff', 'aux'):
                # exiv2 will already have supplied the equivalent Exif tag
                continue
            s_datum = self._xmpData[tag]
            s_datum.setValue(o_datum.value())
